#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define REPEAT 1

__global__ void arrayFunc(float* d_idata, float* d_jdata, float* d_odata, int size)
{
  int tid =  blockDim.x * blockIdx.x + threadIdx.x; 
  if (tid < size) {
     for(int i=0; i < REPEAT; i++)
        d_odata[tid] = d_idata[tid] * __expf(d_jdata[tid]);
  }
}


void initArrayData(float * array, float alpha, int size);
void arrayFuncCPU(const float* h_idata, const float* h_jdata, float* h_odata, int size);
void getChunckInfo(int gpuID, int i, int *d_offset, int *chunk_size, int *h_offset, int *chunk_stream, int nSize, int chunk_size_max, int num_chunk, int num_streams, int gpu_number);

#define NSIZE 2097152
#define CHUNKSIZEMAX 65536
#define NUMSTREAMS 8

#define NGPU 2

int 
main (void) {
  
  float *h_a, *h_b, *h_c;
//-- inserire codice CUDA ----------------
  // definizione dei buffer di memoria
  float *d_a[NGPU], *d_b[NGPU], *d_c[NGPU];
//----------------------------------------
  
  int nsize = NSIZE;
  int nThreads = 256;
  int nBlocks;

  cudaEvent_t start, end;
  float eventEtime;

  int chunk_size_max = CHUNKSIZEMAX;
  int num_streams = NUMSTREAMS;
  int num_chunk;
  int i;
  int h_offset, d_offset;
  int chunk_size, chunk_stream;

//-- inserire codice CUDA ----------------
  // definizione degli stream
  cudaStream_t streams[NGPU][NUMSTREAMS];
//----------------------------------------

  int gpuID, gpu_number;


  // allocazione e inizializzazione dei buffer di memoria su CPU
  cudaMallocHost((void**)&h_a, nsize * sizeof(float));
  cudaMallocHost((void**)&h_b, nsize * sizeof(float));
  cudaMallocHost((void**)&h_c, nsize * sizeof(float));

  initArrayData(h_a, 1.0f, nsize);
  initArrayData(h_b, 10.0f, nsize);


  cudaGetDeviceCount( &gpu_number );
  // calcolo del numero di chunk
  num_chunk = (nsize/gpu_number-1) / chunk_size_max + 1;

  printf("Numero di gpu usate: %d\n", gpu_number);
  printf("Numero di elementi:  %d\n", nsize);
  printf("Numero di streams:   %d\n", num_streams);
  printf("Numero di chunck elaborati da ciascuna GPU: %d\n", num_chunk);

  for (gpuID=0; gpuID<gpu_number; gpuID++) {

//-- inserire codice CUDA ----------------
// selezionare la GPU
    cudaSetDevice(gpuID);
// allocazione dei buffer di memoria su GPU
    cudaMalloc((void**)&d_a[gpuID], num_streams * chunk_size_max * sizeof(float));
    cudaMalloc((void**)&d_b[gpuID], num_streams * chunk_size_max * sizeof(float));
    cudaMalloc((void**)&d_c[gpuID], num_streams * chunk_size_max * sizeof(float));

// creazione degli stream
    for (i = 0; i< num_streams; i++)
      cudaStreamCreate(&streams[gpuID][i]);
//----------------------------------------

  }

  // creazione dei cuda event: start, end
  cudaSetDevice(0);
  cudaEventCreate(&start);  
  cudaEventCreate(&end);

  printf ("\nEseguo il calcolo su GPU ... ");

  cudaEventRecord(start,0);  

  for (gpuID = 0; gpuID < gpu_number; gpuID++) {

    cudaSetDevice(gpuID);

    for (i = 0; i < num_chunk; i++) {

      getChunckInfo(gpuID, i, &d_offset, &chunk_size, &h_offset, &chunk_stream, nsize, chunk_size_max, num_chunk, num_streams, gpu_number);

//-- inserire codice CUDA ----------------
      // copia dei buffer da host a device
      cudaMemcpyAsync(d_a[gpuID]+d_offset, h_a+h_offset, chunk_size*sizeof(float), cudaMemcpyHostToDevice, streams[gpuID][chunk_stream]);
      cudaMemcpyAsync(d_b[gpuID]+d_offset, h_b+h_offset, chunk_size*sizeof(float), cudaMemcpyHostToDevice, streams[gpuID][chunk_stream]);
//----------------------------------------

      // calcolo del numero di blocchi
      nBlocks = (chunk_size-1) / nThreads + 1;

//-- inserire codice CUDA ----------------
      // lancio del kernel arrayFunc
      arrayFunc<<<nBlocks,nThreads, 0, streams[gpuID][chunk_stream]>>>(d_a[gpuID]+d_offset, d_b[gpuID]+d_offset, d_c[gpuID]+d_offset, chunk_size);
//----------------------------------------

//-- inserire codice CUDA ----------------
      // copia del risultato da device a host
      cudaMemcpyAsync(h_c+h_offset, d_c[gpuID]+d_offset, chunk_size*sizeof(float), cudaMemcpyDeviceToHost, streams[gpuID][chunk_stream]);
//----------------------------------------

    }
  }
  
  for (gpuID = 0; gpuID < gpu_number; gpuID++) {
    cudaSetDevice(gpuID);
    cudaDeviceSynchronize();
  }

  cudaSetDevice(0);
  cudaEventRecord(end,0);  
  cudaEventSynchronize(end);
  cudaEventElapsedTime(&eventEtime, start, end);

  printf ("ok\n");

  printf("Tempo impiegato su GPU: %.2f ms\n", eventEtime);

  // calcolo su CPU
  printf("\nEseguo il calcolo su CPU ... ");
  float *cpuResult;
  float eventTimeCPU;
  cudaMallocHost((void**)&cpuResult, nsize * sizeof(float));
  cudaEventRecord(start,0);

  arrayFuncCPU(h_a, h_b, cpuResult, nsize);

  cudaEventRecord(end,0);
  cudaEventElapsedTime(&eventTimeCPU, start, end);
  printf ("ok\n");
  printf("Tempo impiegato su CPU: %.2f ms\n", eventTimeCPU);
  printf("\nSpeed UP CPU/GPU %.1fx\n", eventTimeCPU/eventEtime);
  
  printf("\nControllo il risultato:\n");
  printf ("h_c[0]       = %f\n", h_c[0]);
  printf ("cpuResult[0] = %f\n", cpuResult[0]);


  // liberare le risorse sul device
  for (gpuID = 0; gpuID < gpu_number; gpuID++) {
    cudaSetDevice(gpuID);
    for (i = 0; i< num_streams; i++)
      cudaStreamDestroy(streams[gpuID][i]);
  }
  cudaSetDevice(0);
  cudaEventDestroy(start);
  cudaEventDestroy(end);

  // liberare le risorse sull'host
  cudaFreeHost(h_a);
  cudaFreeHost(h_b);
  cudaFreeHost(h_c);

  return 0;

}


void 
initArrayData(float * array, float alpha, int size)
{ 
  int i;
  for (i=0; i< size; i++) 
    array[i] = alpha * (float) rand() / (float) RAND_MAX;
}

void getChunckInfo(int gpuID, int i, int *d_offset, int *chunk_size, int *h_offset, int *chunk_stream, int nSize, int chunk_size_max, int num_chunk, int num_streams, int gpu_number){

  int Reminder = nSize%chunk_size_max;

  *h_offset = i*chunk_size_max + gpuID * (nSize/gpu_number);
  *chunk_stream = i%num_streams;
  *chunk_size   = chunk_size_max;
  *d_offset = *chunk_stream * chunk_size_max;

  if (Reminder && (i == num_chunk-1)) *chunk_size = Reminder;

}

void arrayFuncCPU(const float* h_idata, const float* h_jdata, float* h_odata, int size)
{
  int i, j; 
  for (i=0; i<size; i++)
     for(j=0; j<REPEAT; j++)
        h_odata[i] = h_idata[i] * expf(h_jdata[i]);
}
